# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math
from typing import Any, Dict, List, Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from fairseq import options, utils
from fairseq.models import (
    FairseqEncoder,
    FairseqEncoderDecoderModel,
    FairseqIncrementalDecoder,
    register_model,
    register_model_architecture,
)
from fairseq.models.fairseq_encoder import EncoderOut
from fairseq.modules import (
    AdaptiveSoftmax,
    LayerNorm,
    PositionalEmbedding,
    SinusoidalPositionalEmbedding,
    TransformerDecoderLayer,
    TransformerEncoderLayer,
)
from torch import Tensor

DEFAULT_MAX_SOURCE_POSITIONS = 1024
DEFAULT_MAX_TARGET_POSITIONS = 1024


@register_model("transformer")
class TransformerModel(FairseqEncoderDecoderModel):
    """
    Transformer model from `"Attention Is All You Need" (Vaswani, et al, 2017)
    <https://arxiv.org/abs/1706.03762>`_.

    Args:
        encoder (TransformerEncoder): the encoder
        decoder (TransformerDecoder): the decoder

    The Transformer model provides the following named architectures and
    command-line arguments:

    .. configargparse::
        :ref: fairseq.models.transformer_parser
        :prog:
    """

    @classmethod
    def hub_models(cls):
        # fmt: off

        def moses_subword(path):
            return {
                'path': path,
                'tokenizer': 'moses',
                'bpe': 'subword_nmt',
            }

        def moses_fastbpe(path):
            return {
                'path': path,
                'tokenizer': 'moses',
                'bpe': 'fastbpe',
            }

        return {
            'transformer.wmt14.en-fr': moses_subword('https://dl.fbaipublicfiles.com/fairseq/models/wmt14.en-fr.joined-dict.transformer.tar.bz2'),
            'transformer.wmt16.en-de': 'https://dl.fbaipublicfiles.com/fairseq/models/wmt16.en-de.joined-dict.transformer.tar.bz2',
            'transformer.wmt18.en-de': moses_subword('https://dl.fbaipublicfiles.com/fairseq/models/wmt18.en-de.ensemble.tar.gz'),
            'transformer.wmt19.en-de': moses_fastbpe('https://dl.fbaipublicfiles.com/fairseq/models/wmt19.en-de.joined-dict.ensemble.tar.gz'),
            'transformer.wmt19.en-ru': moses_fastbpe('https://dl.fbaipublicfiles.com/fairseq/models/wmt19.en-ru.ensemble.tar.gz'),
            'transformer.wmt19.de-en': moses_fastbpe('https://dl.fbaipublicfiles.com/fairseq/models/wmt19.de-en.joined-dict.ensemble.tar.gz'),
            'transformer.wmt19.ru-en': moses_fastbpe('https://dl.fbaipublicfiles.com/fairseq/models/wmt19.ru-en.ensemble.tar.gz'),
            'transformer.wmt19.en-de.single_model': moses_fastbpe('https://dl.fbaipublicfiles.com/fairseq/models/wmt19.en-de.joined-dict.single_model.tar.gz'),
            'transformer.wmt19.en-ru.single_model': moses_fastbpe('https://dl.fbaipublicfiles.com/fairseq/models/wmt19.en-ru.single_model.tar.gz'),
            'transformer.wmt19.de-en.single_model': moses_fastbpe('https://dl.fbaipublicfiles.com/fairseq/models/wmt19.de-en.joined-dict.single_model.tar.gz'),
            'transformer.wmt19.ru-en.single_model': moses_fastbpe('https://dl.fbaipublicfiles.com/fairseq/models/wmt19.ru-en.single_model.tar.gz'),
        }
        # fmt: on

    def __init__(self, args, encoder, decoder, pooler):
        super().__init__(encoder, decoder, pooler)
        self.args = args
        encoder.global_positions = args.global_positions

        self.supports_align_args = True
        self.handles = {'pool_debug_file': open(self.args.pool_debug_file,
                                                'a+') if self.args.pool_debug_file else None}



    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument('--activation-fn',
                            choices=utils.get_available_activation_fns(),
                            help='activation function to use')
        parser.add_argument('--dropout', type=float, metavar='D',
                            help='dropout probability')
        parser.add_argument('--attention-dropout', type=float, metavar='D',
                            help='dropout probability for attention weights')
        parser.add_argument('--activation-dropout', '--relu-dropout', type=float, metavar='D',
                            help='dropout probability after activation in FFN.')
        parser.add_argument('--encoder-embed-path', type=str, metavar='STR',
                            help='path to pre-trained encoder embedding')
        parser.add_argument('--encoder-embed-dim', type=int, metavar='N',
                            help='encoder embedding dimension')
        parser.add_argument('--encoder-ffn-embed-dim', type=int, metavar='N',
                            help='encoder embedding dimension for FFN')
        parser.add_argument('--encoder-layers', type=int, metavar='N',
                            help='num encoder layers')
        parser.add_argument('--encoder-attention-heads', type=int, metavar='N',
                            help='num encoder attention heads')
        parser.add_argument('--encoder-normalize-before', action='store_true',
                            help='apply layernorm before each encoder block')
        parser.add_argument('--encoder-learned-pos', action='store_true',
                            help='use learned positional embeddings in the encoder')
        parser.add_argument('--decoder-embed-path', type=str, metavar='STR',
                            help='path to pre-trained decoder embedding')
        parser.add_argument('--decoder-embed-dim', type=int, metavar='N',
                            help='decoder embedding dimension')
        parser.add_argument('--decoder-ffn-embed-dim', type=int, metavar='N',
                            help='decoder embedding dimension for FFN')
        parser.add_argument('--decoder-layers', type=int, metavar='N',
                            help='num decoder layers')
        parser.add_argument('--decoder-attention-heads', type=int, metavar='N',
                            help='num decoder attention heads')
        parser.add_argument('--decoder-learned-pos', action='store_true',
                            help='use learned positional embeddings in the decoder')
        parser.add_argument('--decoder-normalize-before', action='store_true',
                            help='apply layernorm before each decoder block')
        parser.add_argument('--share-decoder-input-output-embed', action='store_true',
                            help='share decoder input and output embeddings')
        parser.add_argument('--share-all-embeddings', action='store_true',
                            help='share encoder, decoder and output embeddings'
                                 ' (requires shared dictionary and embed dim)')
        parser.add_argument('--no-token-positional-embeddings', default=False, action='store_true',
                            help='if set, disables positional embeddings (outside self attention)')

        parser.add_argument('--adaptive-softmax-cutoff', metavar='EXPR',
                            help='comma separated list of adaptive softmax cutoff points. '
                                 'Must be used with adaptive_loss criterion'),
        parser.add_argument('--adaptive-softmax-dropout', type=float, metavar='D',
                            help='sets adaptive softmax dropout for the tail projections')
        # args for "Cross+Self-Attention for Transformer Models" (Peitz et al., 2019)
        parser.add_argument('--no-cross-attention', default=False, action='store_true',
                            help='do not perform cross-attention')
        parser.add_argument('--cross-self-attention', default=False, action='store_true',
                            help='perform cross+self-attention')
        parser.add_argument('--layer-wise-attention', default=False, action='store_true',
                            help='perform layer-wise attention (cross-attention or cross+self-attention)')
        # args for "Reducing Transformer Depth on Demand with Structured Dropout" (Fan et al., 2019)
        parser.add_argument('--encoder-layerdrop', type=float, metavar='D', default=0,
                            help='LayerDrop probability for encoder')
        parser.add_argument('--decoder-layerdrop', type=float, metavar='D', default=0,
                            help='LayerDrop probability for decoder')
        parser.add_argument('--encoder-layers-to-keep', default=None,
                            help='which layers to *keep* when pruning as a comma-separated list')
        parser.add_argument('--decoder-layers-to-keep', default=None,
                            help='which layers to *keep* when pruning as a comma-separated list')
        parser.add_argument('--layernorm-embedding', action='store_true',
                            help='add layernorm to embedding')
        parser.add_argument('--no-scale-embedding', action='store_true',
                            help='if True, dont scale embeddings')
        #  POOLER ARGS
        parser.add_argument('--no-decoder-positional-embeddings', default=False, action='store_true',
                            help='if set, disables positional embeddings in decoder')
        parser.add_argument('--encoder-pooling', type=str, default='lambda',
                            choices=['lambda', 'topk', 'threshold', 'regression', 'sinkhorn',
                                     'nonet'],
                            help='name of the pooling mechanism to use between encoder ' \
                                 'and decoder (no pooling by default)')
        parser.add_argument('--encoder-pooling-arch', type=str, default='linear',
                            choices=['linear', 'ffn', 'attn', 'kth', 'random', 'zeroth_dim',
                                     'power_bert', 'mean_pool', 'max_pool'],
                            help='pooling classifier architecture')
        parser.add_argument('--pooled-length', type=int, default=512,
                            help='length of sequence passed to decoder')
        parser.add_argument('--pool-debug-file', type=str, default='',
                            help='file to save debug output')
        parser.add_argument('--pool-use-cut-embedding', action='store_true',
                            help='add cut embedding to pooled selected indices')
        parser.add_argument('--pool-use-bias', action='store_true',
                            help='use index-based bias')
        parser.add_argument('--pooler-to-decoder', action='store_true',
                            help='use zero-depth encoder (only embedder) and pool '
                                 'straight to the decoder')
        parser.add_argument('--global-positions', type=int, default=0,
                            help='add position to each token based on their chunk position')
        parser.add_argument('--softmax-n', type=int, default=2,
                            help='How many times to perform softmax in a series for '
                                 'normalizing scores in the NoNet')
        parser.add_argument('--speed-up-const', type=float, default=1.0,
                            help='Multiplier, that speeds up softmax normalization,'
                                 ' but is numerically unstable for fp16.')
        parser.add_argument('--flip-right', type=int, default=1,
                            help='Whether the right side of scores that will be softmaxed against'
                                 'top-k scores should be `mirrored` left-to-right. '
                                 'This should be preferred IMO')
        parser.add_argument('--chunk-size', type=int, default=512,
                            help='Size of the chunk, that will be consumed by encoder at once. '
                                 'Note that memory depends quadratically over `chunk size` '
                                 'and linearly over '
                                 'number of chunks one will get after splitting into chunks.')
        parser.add_argument('--apply-hierarchical-pooling', type=int, default=0,
                            help='Whether to use hierarchical pooling for each layer: '
                                 '(unfold, encode, fold back and pool) in a predefined encoder '
                                 'layers or do basic stuff for the whole encoder.')
        parser.add_argument('--enc-layers-and-token-width', type=str,
                            default="{0: 2048, 1: 1024, 2: 256}",
                            help='List of two-tuples, where each two-tuple contains layer number '
                                 'and strength of bottleneck to be applied (e.g. (1, 2) means there will'
                                 ' be narrowed down from x to x/2 chunks after the first '
                                 'layer in the encoder).')
        parser.add_argument('--sort-back', type=int, default=0,
                            help='Whether to sort-back pooled representations at each step.'
                                 ' This used to be 0 for enc-pool-dec architectures, '
                                 'but in hierarchical sorting, there is a need to sort I believe.')
        parser.add_argument('--pooler-iterative-topk', type=int, default=0,
                            help='Whether to use iterative topk approach. '
                                 'The number specified here controls the alpha speed up const for '
                                 'iterative solutions from https://arxiv.org/pdf/1708.00111.pdf')
        parser.add_argument('--pool-reuse-scorer', type=int, default=1,
                            help='Whether to reuse the same scorer at every layer or build new '
                                 'linear/ffn projections')
        parser.add_argument('--log-time-every-n-steps', type=int, default=1,
                            help='How often to save time-speed logs into .log file '
                                 '(next to config).')
        # ---- PEGASUS SPECIFIC ----
        parser.add_argument('--use-pegasus', type=int, default=0,
                            help='Whether to use pegasus loss or normal MLM')
        parser.add_argument('--pegasus-gsr', type=float, default=0.15,
                            help='Pegasus Gap-to-Sentence Ratio. (typically should be 15, 30, 45%')
        parser.add_argument('--pooler-remove-corrupted-datapoint', type=int, default=1,
                            help='For arxiv dataset, a part of the input has longer abstract than '
                                 'body, which was the result of automatic parsing and is in the '
                                 'original arxiv dataset.')
        # ---- SPARSE ATTENTION SPECIFIC ----
        parser.add_argument('--use-sparse-attn', type=str, default='pooler',
                            choices=['linformer', 'only_blockwise', 'sparse', 'pooler',
                                     'linformer_no_cross_decoder', 'pooler_no_block'],
                            help='type of attention to use. pooler uses multihead,'
                                 ' linformer uses linearmultihead, and sparse uses sparsemultihead ')
        parser.add_argument('--compress', type=int, default=4,
                            help='Compression factor in the linformer attention module')
        parser.add_argument('--funnel', type=int, default=0,
                            help='whether to use funnel transformer (mean avg pooling '
                                 '+ encoder latent upsampling'
                                 '+ residual connection to decoder)')
        parser.add_argument('--residual_takes_from_layer_i', type=int, default=0,
                            help='position of a layer from which we will start residual '
                                 'connection to the decoder')

        # ---- Inference tricks
        parser.add_argument('--hard-topk-inference', action='store_true', default=False,
                            help='whether to infer with hard topk instead of soft selector '
                                 '(happens in no_net class)')

        # fmt: on

    @classmethod
    def build_model(cls, args, task):
        """Build a new model instance."""

        # make sure all arguments are present in older models
        base_architecture(args)

        if args.encoder_layers_to_keep:
            args.encoder_layers = len(args.encoder_layers_to_keep.split(","))
        if args.decoder_layers_to_keep:
            args.decoder_layers = len(args.decoder_layers_to_keep.split(","))

        if getattr(args, "max_source_positions", None) is None:
            args.max_source_positions = DEFAULT_MAX_SOURCE_POSITIONS
        if getattr(args, "max_target_positions", None) is None:
            args.max_target_positions = DEFAULT_MAX_TARGET_POSITIONS

        src_dict, tgt_dict = task.source_dictionary, task.target_dictionary

        def build_embedding(dictionary, embed_dim, path=None):
            num_embeddings = len(dictionary)
            padding_idx = dictionary.pad()
            emb = Embedding(num_embeddings, embed_dim, padding_idx)
            # if provided, load from preloaded dictionaries
            if path:
                embed_dict = utils.parse_embedding(path)
                utils.load_embedding(embed_dict, dictionary, emb)
            return emb

        if args.share_all_embeddings:
            if src_dict != tgt_dict:
                raise ValueError("--share-all-embeddings requires a joined dictionary")
            if args.encoder_embed_dim != args.decoder_embed_dim:
                raise ValueError(
                    "--share-all-embeddings requires --encoder-embed-dim to match --decoder-embed-dim"
                )
            if args.decoder_embed_path and (
                args.decoder_embed_path != args.encoder_embed_path
            ):
                raise ValueError(
                    "--share-all-embeddings not compatible with --decoder-embed-path"
                )
            encoder_embed_tokens = build_embedding(
                src_dict, args.encoder_embed_dim, args.encoder_embed_path
            )
            decoder_embed_tokens = encoder_embed_tokens
            args.share_decoder_input_output_embed = True
        else:
            encoder_embed_tokens = build_embedding(
                src_dict, args.encoder_embed_dim, args.encoder_embed_path
            )
            decoder_embed_tokens = build_embedding(
                tgt_dict, args.decoder_embed_dim, args.decoder_embed_path
            )
        pooler = cls.build_pooler(args)
        encoder = cls.build_encoder(args, src_dict, encoder_embed_tokens, pooler)
        decoder = cls.build_decoder(args, tgt_dict, decoder_embed_tokens)


        return cls(args, encoder, decoder, None)

    @classmethod
    def build_encoder(cls, args, src_dict, embed_tokens, pooler):
        return TransformerEncoder(args, src_dict, embed_tokens, pooler)

    @classmethod
    def build_pooler(cls, args):
        # FIXME: this is ugly, but it didn't work up there
        from fairseq.models import SkynetPooler
        return SkynetPooler(args)

    @classmethod
    def build_decoder(cls, args, tgt_dict, embed_tokens):
        return TransformerDecoder(
            args,
            tgt_dict,
            embed_tokens,
            no_encoder_attn=getattr(args, "no_cross_attention", False),
        )

    # TorchScript doesn't support optional arguments with variable length (**kwargs).
    # Current workaround is to add union of all arguments in child classes.
    def forward(
        self,
        src_tokens,
        src_lengths,
        prev_output_tokens,
        cls_input: Optional[Tensor] = None,
        return_all_hiddens: bool = True,
        features_only: bool = False,
        alignment_layer: Optional[int] = None,
        alignment_heads: Optional[int] = None,
    ):
        """
        Run the forward pass for an encoder-decoder model.

        Copied from the base class, but without ``**kwargs``,
        which are not supported by TorchScript.
        """
        encoder_out = self.encoder(
            src_tokens,
            src_lengths=src_lengths,
            cls_input=cls_input,
            return_all_hiddens=return_all_hiddens,
        )
        eosh = encoder_out.encoder_out.shape
        if self.pooler:
            assert eosh[0] == self.args.pooled_length
        #self.write_to_debug_maybe(self.handles, self.encoder, src_tokens_padded, selected_indices)
        if self.args.funnel:
            """upsample and residual"""
            num_layer_for_residual_in_encoder = self.args.residual_takes_from_layer_i
            residual_representation = encoder_out.encoder_states[num_layer_for_residual_in_encoder]
            upsample_factor = residual_representation.shape[0] // encoder_out.encoder_out.shape[0]
            upsampled_latent = torch.repeat_interleave(encoder_out.encoder_out, upsample_factor, 0)
            funnel_out = residual_representation + upsampled_latent

            funnel_padding_mask = torch.repeat_interleave(encoder_out.encoder_padding_mask, upsample_factor, 1)
            funnel_encoder_out = EncoderOut(
                encoder_out=funnel_out,  # T x B x C
                encoder_padding_mask=funnel_padding_mask,  # B x T
                encoder_embedding=None,  # encoder_out.encoder_embedding,  # B x T x C
                encoder_states=None,  # encoder_out.encoder_states,  # List[T x B x C]
                encoder_last_attn_matrix=None,
            )
            encoder_out = funnel_encoder_out

        decoder_out = self.decoder(
            prev_output_tokens,
            encoder_out=encoder_out,
            features_only=features_only,
            alignment_layer=alignment_layer,
            alignment_heads=alignment_heads,
            src_lengths=src_lengths,
            return_all_hiddens=return_all_hiddens,
        )
        return decoder_out

    # Since get_normalized_probs is in the Fairseq Model which is not scriptable,
    # I rewrite the get_normalized_probs from Base Class to call the
    # helper function in the Base Class.
    @torch.jit.export
    def get_normalized_probs(
        self,
        net_output: Tuple[Tensor, Dict[str, List[Optional[Tensor]]]],
        log_probs: bool,
        sample: Optional[Dict[str, Tensor]] = None,
    ):
        """Get normalized probabilities (or log probs) from a net's output."""
        return self.get_normalized_probs_scriptable(net_output, log_probs, sample)


@register_model("transformer_align")
class TransformerAlignModel(TransformerModel):
    """
    See "Jointly Learning to Align and Translate with Transformer
    Models" (Garg et al., EMNLP 2019).
    """

    def __init__(self, encoder, decoder, args):
        super().__init__(args, encoder, decoder)
        self.alignment_heads = args.alignment_heads
        self.alignment_layer = args.alignment_layer
        self.full_context_alignment = args.full_context_alignment

    @staticmethod
    def add_args(parser):
        # fmt: off
        super(TransformerAlignModel, TransformerAlignModel).add_args(parser)
        parser.add_argument('--alignment-heads', type=int, metavar='D',
                            help='Number of cross attention heads per layer to supervised with alignments')
        parser.add_argument('--alignment-layer', type=int, metavar='D',
                            help='Layer number which has to be supervised. 0 corresponding to the bottommost layer.')
        parser.add_argument('--full-context-alignment', type=bool, metavar='D',
                            help='Whether or not alignment is supervised conditioned on the full target context.')
        # fmt: on

    @classmethod
    def build_model(cls, args, task):
        # set any default arguments
        transformer_align(args)

        transformer_model = TransformerModel.build_model(args, task)
        return TransformerAlignModel(
            transformer_model.encoder, transformer_model.decoder, args
        )

    def forward(self, src_tokens, src_lengths, prev_output_tokens):
        encoder_out = self.encoder(src_tokens, src_lengths)
        return self.forward_decoder(prev_output_tokens, encoder_out)

    def forward_decoder(
        self,
        prev_output_tokens,
        encoder_out=None,
        incremental_state=None,
        features_only=False,
        **extra_args,
    ):
        attn_args = {
            "alignment_layer": self.alignment_layer,
            "alignment_heads": self.alignment_heads,
        }
        decoder_out = self.decoder(prev_output_tokens, encoder_out, **attn_args)

        if self.full_context_alignment:
            attn_args["full_context_alignment"] = self.full_context_alignment
            _, alignment_out = self.decoder(
                prev_output_tokens,
                encoder_out,
                features_only=True,
                **attn_args,
                **extra_args,
            )
            decoder_out[1]["attn"] = alignment_out["attn"]

        return decoder_out


class TransformerEncoder(FairseqEncoder):
    """
    Transformer encoder consisting of *args.encoder_layers* layers. Each layer
    is a :class:`TransformerEncoderLayer`.

    Args:
        args (configargparse.Namespace): parsed command-line arguments
        dictionary (~fairseq.data.Dictionary): encoding dictionary
        embed_tokens (torch.nn.Embedding): input embedding
    """

    def __init__(self, args, dictionary, embed_tokens, pooler):
        super().__init__(dictionary)
        self.pooler = pooler
        self.register_buffer("version", torch.Tensor([3]))
        self.args = args
        self.dropout = args.dropout
        self.encoder_layerdrop = args.encoder_layerdrop

        embed_dim = embed_tokens.embedding_dim
        self.padding_idx = embed_tokens.padding_idx
        self.max_source_positions = args.max_source_positions

        self.embed_tokens = embed_tokens

        self.embed_scale = 1.0 if args.no_scale_embedding else math.sqrt(embed_dim)

        self.embed_positions = (
            PositionalEmbedding(
                args.max_source_positions,
                embed_dim,
                self.padding_idx,
                learned=args.encoder_learned_pos,
                args=args,
            )
            if not args.no_token_positional_embeddings
            else None
        )

        self.layer_wise_attention = getattr(args, "layer_wise_attention", False)

        self.layers = nn.ModuleList([])
        self.layers.extend(
            [TransformerEncoderLayer(args) for i in range(args.encoder_layers)]
        )
        self.num_layers = len(self.layers)

        if args.encoder_normalize_before:
            self.layer_norm = LayerNorm(embed_dim)
        else:
            self.layer_norm = None
        if getattr(args, "layernorm_embedding", False):
            self.layernorm_embedding = LayerNorm(embed_dim)
        else:
            self.layernorm_embedding = None

        from fairseq.models.pooler_utils import (unfold_maybe,
                                                 fold_back_maybe,
                                                 write_to_debug_maybe,
                                                 )

        self.unfold_maybe = unfold_maybe
        self.fold_back_maybe = fold_back_maybe
        self.write_to_debug_maybe = write_to_debug_maybe


    def forward_embedding(self, src_tokens):
        # embed tokens and positions
        x = embed = self.embed_scale * self.embed_tokens(src_tokens)
        if self.embed_positions is not None:
            x = embed + self.embed_positions(src_tokens, chunks_num=self.chunks_num)
        if self.layernorm_embedding is not None:
            x = self.layernorm_embedding(x)
        x = F.dropout(x, p=self.dropout, training=self.training)
        return x, embed

    def forward(
        self,
        src_tokens,
        src_lengths,
        cls_input: Optional[Tensor] = None,
        return_all_hiddens: bool = False,
        need_head_weights: bool = False,
    ):
        """
        Args:
            src_tokens (LongTensor): tokens in the source language of shape
                `(batch, src_len)`
            src_lengths (torch.LongTensor): lengths of each source sentence of
                shape `(batch)`
            return_all_hiddens (bool, optional): also return all of the
                intermediate hidden states (default: False).

        Returns:
            namedtuple:
                - **encoder_out** (Tensor): the last encoder layer's output of
                  shape `(src_len, batch, embed_dim)`
                - **encoder_padding_mask** (ByteTensor): the positions of
                  padding elements of shape `(batch, src_len)`
                - **encoder_embedding** (Tensor): the (scaled) embedding lookup
                  of shape `(batch, src_len, embed_dim)`
                - **encoder_states** (List[Tensor]): all intermediate
                  hidden states of shape `(src_len, batch, embed_dim)`.
                  Only populated if *return_all_hiddens* is True.
        """
        need_attn = True if self.pooler.scorer in ['attn', 'power_bert'] else False

        src_lengths_old = src_lengths
        src_lengths, src_tokens_padded, src_tokens, old_bs, chunks_num \
            = self.unfold_maybe(self.pooler, self.args.chunk_size, self, src_lengths, src_tokens)
        self.chunks_num = chunks_num

        encoder_out = self.forward_internal(need_attn, return_all_hiddens, src_tokens, src_lengths_old,
                                            old_bs)

        if not self.args.apply_hierarchical_pooling:
            encoder_out, attn_scores = self.fold_back_maybe(self.pooler, self.args, encoder_out,
                                                       old_bs, need_attn)
            # encoder_out : Len x Batch x Dim
            cut_embedding = None
            if hasattr(self.args, 'pool_use_cut_embedding'):
                cut_embedding = self.encoder.embed_tokens._parameters['weight'][
                    self.dictionary.cut_index]

            shp = encoder_out[0].shape
            is_blockwise = getattr(self.args, 'use_sparse_attn', 'none') == 'only_blockwise'
            stop_blockwise = getattr(self.pooler.args, 'use_sparse_attn',
                                     'none') == 'pooler_no_block'  # 'pooler_no_block' TURNS OFF BLOCKWISE ATTN
            stop_blockwise |= getattr(self.pooler.args, 'use_sparse_attn', 'none') == 'linformer'
            stop_blockwise |= getattr(self.pooler.args, 'use_sparse_attn',
                                      'none') == 'linformer_no_cross_decoder'
            if (not self.pooler.is_lambda and not stop_blockwise) or is_blockwise:
                assert shp[0] == src_tokens_padded.shape[1]
                assert shp[1] == old_bs == src_tokens_padded.shape[0]
                assert shp[2] == self.args.encoder_embed_dim
            else:
                assert shp[0] == src_tokens.shape[1]
                assert shp[1] == old_bs == src_tokens.shape[0]
                assert shp[2] == self.args.encoder_embed_dim

            encoder_out, selected_indices = self.pooler(encoder_out,
                                                        src_lengths=src_lengths_old,
                                                        cut_embedding=cut_embedding,
                                                        attn_scores=attn_scores,
                                                        pooled_length=0 if self.pooler.is_lambda
                                                        else self.pooler.selector.pooled_len,
                                                        layer_i=self.args.encoder_layers)
            assert encoder_out[0].shape[1:] == shp[1:]
        return encoder_out

    def forward_internal(self, need_head_weights, return_all_hiddens, src_tokens, src_lengths_old,
                         old_bs):
        """The real encodding, embedding, etc. """

        if self.layer_wise_attention:
            return_all_hiddens = True
        x, encoder_embedding = self.forward_embedding(src_tokens)
        # B x T x C -> T x B x C
        x = x.transpose(0, 1)
        # compute padding mask
        encoder_padding_mask = src_tokens.eq(self.padding_idx)
        encoder_states = [] if return_all_hiddens else None
        hidden_dim = x.shape[-1]
        # # pad input to the  `max_source_positions`
        # if self.args.apply_hierarchical_pooling:
        #     old_bs = x.shape[1]
        #     hidden_dim = x.shape[-1]
        #     encoder_padding_mask, x = self.pad_x_to_max_source_positions(encoder_padding_mask,
        #                                                                  hidden_dim, old_bs, x)
        # encoder layers
        for layer_i, layer in enumerate(self.layers):
            # add LayerDrop (see https://arxiv.org/abs/1909.11556 for description)
            dropout_probability = torch.empty(1).uniform_()
            if not self.training or (dropout_probability > self.encoder_layerdrop):
                # TODO: unfold here
                if self.args.apply_hierarchical_pooling:
                    self.prepare_hierarchical_pooler(layer_i)

                    if layer_i > 0:
                        chunk_size = min(x.shape[0], self.args.chunk_size)

                        self.chunks_num = x.shape[0] // chunk_size
                        src_lengths_old = torch.LongTensor([x.shape[0]] * old_bs).to(x.device)

                        x = x.unfold(0, chunk_size, chunk_size)\
                            .permute(1, 0, 3, 2)\
                            .flatten(0, 1)\
                            .transpose(0, 1)  # Len x Batch x Dim

                        assert x.shape[1] == old_bs * self.chunks_num

                        # Recreate False-filled mask for layers after pooling,
                        # as this is not tractable, and can be assumed without loss of generality
                        encoder_padding_mask = torch.zeros(x.shape[1], x.shape[0]).\
                            to(encoder_padding_mask.device).eq(1)

                x, attn_matrix = layer(x, encoder_padding_mask.to(x.device),
                                       need_head_weights=need_head_weights)
                # TODO: fold back here
                if self.args.apply_hierarchical_pooling:
                    # unfold, but keep:   Len x Batch x Dim
                    # x = x \
                    #     .transpose(0, 1) \
                    #     .reshape([old_bs, -1, hidden_dim]) \
                    #     .transpose(0, 1)

                    encoder_out = EncoderOut(
                        encoder_out=x,  # T x B x C
                        encoder_padding_mask=encoder_padding_mask,  # B x T
                        encoder_embedding=encoder_embedding,  # B x T x C
                        encoder_states=encoder_states,  # List[T x B x C]
                        encoder_last_attn_matrix=attn_matrix if len(self.layers) else None,
                        # B x T x T
                    )
                    encoder_out, attn_scores = self.fold_back_maybe(self.pooler, self.args, encoder_out, old_bs, need_head_weights)
                    shp = encoder_out[0].shape
                    encoder_out, selected_indices = self.pooler(encoder_out,
                                                                src_lengths=src_lengths_old,
                                                                cut_embedding=False,
                                                                attn_scores=attn_scores,
                                                                pooled_length=self.pooler.selector.pooled_len,
                                                                layer_i=layer_i)
                    assert encoder_out[0].shape[1:] == shp[1:]
                    x = encoder_out.encoder_out

                if return_all_hiddens:
                    assert encoder_states is not None
                    encoder_states.append(x)

        if self.layer_norm is not None:
            x = self.layer_norm(x)
            if return_all_hiddens:
                encoder_states[-1] = x
        # There may be a tricky case, when `chunk_size` is not same as targeted `pooled_length`
        # if self.args.pooled_length == self.args.chunk_size:
        #     encoder_padding_mask = torch.zeros(x.shape[1], x.shape[0]). \
        #         to(encoder_padding_mask.device).eq(1)
        #     assert encoder_padding_mask.shape[1] == x.shape[0]
        #     assert encoder_padding_mask.shape[0] == x.shape[1]
        # else:
        if self.args.apply_hierarchical_pooling:
            encoder_padding_mask = torch.zeros((x.shape[1], x.shape[0]),
                dtype=encoder_padding_mask.dtype,
                device=encoder_padding_mask.device).eq(1)

        return EncoderOut(
            encoder_out=x,  # T x B x C
            encoder_padding_mask=encoder_padding_mask,  # B x T
            encoder_embedding=encoder_embedding,  # B x T x C
            encoder_states=encoder_states,  # List[T x B x C]
            encoder_last_attn_matrix=attn_matrix if len(self.layers) else None,  # B x T x T
        )

    def prepare_hierarchical_pooler(self, layer_i):
        """Build pyramid for hierarchical pooler"""
        # set it before pooling for the current layer
        if self.args.enc_layers_and_token_width:
            if not hasattr(self, 'pooler_pyramid'):
                self.build_pooler_pyramid()
            layer_settings = self.pooler_pyramid[layer_i]
            self.pooler.pooler_config.input_len = layer_settings['input_len']
            self.pooler.pooler_config.pooled_len = layer_settings['pooled_len']
            self.pooler.pooler_config.depth = -1        # NoNet can derive it on its own.
            self.pooler.pooler_config.sort_back = self.args.sort_back
        else:
            """Default setting is to divide by 2 at each layer(step)"""
            if layer_i == 0:
                # initialize ...
                self.pooler.pooler_config.input_len = self.args.max_source_positions
                self.pooler.pooler_config.pooled_len = self.args.max_source_positions // 2
                self.pooler.pooler_config.sort_back = self.args.sort_back
            else:
                #  ... and reuse the same pooler
                self.pooler.pooler_config.input_len //= 2
                self.pooler.pooler_config.pooled_len //= 2
        # always set config to the selector
        self.pooler.selector.set_config(self.pooler.pooler_config)

    def build_pooler_pyramid(self):
        targets = eval(self.args.enc_layers_and_token_width)
        max_layer = len(self.layers) - 1
        assert max(
            targets.keys()) <= max_layer,\
            "Wrongly constructed targets! " \
            "Too many layers mentioned in the targets of 'enc_layers_and_token_width', " \
            "but not enought encoder layers!"
        pyramid = {}
        input_len = self.args.max_source_positions
        pooled_len = input_len
        for lay_i in range(max_layer + 1):
            if lay_i in targets:
                pooled_len = targets[lay_i]
            layer_settings = {'input_len': input_len, 'pooled_len': pooled_len}
            input_len = pooled_len
            pyramid[lay_i] = layer_settings
        assert pyramid[
                   max_layer]['pooled_len'] == self.args.pooled_length, \
            f"Wrongly constructed pyramid! Last target is {pyramid[max_layer]}, " \
            f"whereas args.pooled_length got {self.args.pooled_length}"
        for k, v in pyramid.items():
            print(k, ': ', v)
        self.pooler_pyramid = pyramid

    def pad_x_to_max_source_positions(self, encoder_padding_mask, hidden_dim, old_bs, x):
        pad_len = self.args.max_source_positions - x.shape[0]
        # TODO: padding should be with 1s, but 1s vectors should not be processed!
        enc_padder = torch.ones(
            (old_bs, pad_len),
            dtype=encoder_padding_mask.dtype,
            device=encoder_padding_mask.device
        )
        encoder_padding_mask = torch.cat((encoder_padding_mask, enc_padder), dim=1)
        encoder_padding_mask = encoder_padding_mask.unfold(1, self.args.chunk_size,
                                                           self.args.chunk_size).flatten(0, 1)
        doc_padder = torch.zeros((pad_len, old_bs, hidden_dim), dtype=x.dtype, device=x.device)
        x = torch.cat((x, doc_padder), dim=0)
        return encoder_padding_mask, x

    @torch.jit.export
    def reorder_encoder_out(self, encoder_out: EncoderOut, new_order):
        """
        Reorder encoder output according to *new_order*.

        Args:
            encoder_out: output from the ``forward()`` method
            new_order (LongTensor): desired order

        Returns:
            *encoder_out* rearranged according to *new_order*
        """
        new_encoder_out: Dict[str, Tensor] = {}

        new_encoder_out["encoder_out"] = (
            encoder_out.encoder_out
            if encoder_out.encoder_out is None
            else encoder_out.encoder_out.index_select(1, new_order)
        )
        new_encoder_out["encoder_padding_mask"] = (
            encoder_out.encoder_padding_mask
            if encoder_out.encoder_padding_mask is None
            else encoder_out.encoder_padding_mask.index_select(0, new_order)
        )
        new_encoder_out["encoder_embedding"] = (
            encoder_out.encoder_embedding
            if encoder_out.encoder_embedding is None
            else encoder_out.encoder_embedding.index_select(0, new_order)
        )

        encoder_states = encoder_out.encoder_states
        if encoder_states is not None:
            for idx, state in enumerate(encoder_states):
                encoder_states[idx] = state.index_select(1, new_order)

        return EncoderOut(
            encoder_out=new_encoder_out["encoder_out"],  # T x B x C
            encoder_padding_mask=new_encoder_out["encoder_padding_mask"],  # B x T
            encoder_embedding=new_encoder_out["encoder_embedding"],  # B x T x C
            encoder_states=encoder_states,  # List[T x B x C]
            encoder_last_attn_matrix=None,
        )

    def max_positions(self):
        """Maximum input length supported by the encoder."""
        if self.embed_positions is None:
            return self.max_source_positions
        return min(self.max_source_positions, self.embed_positions.max_positions)

    def buffered_future_mask(self, tensor):
        dim = tensor.size(0)
        if (
            not hasattr(self, "_future_mask")
            or self._future_mask is None
            or self._future_mask.device != tensor.device
        ):
            self._future_mask = torch.triu(
                utils.fill_with_neg_inf(tensor.new(dim, dim)), 1
            )
            if self._future_mask.size(0) < dim:
                self._future_mask = torch.triu(
                    utils.fill_with_neg_inf(self._future_mask.resize_(dim, dim)), 1
                )
        return self._future_mask[:dim, :dim]

    def upgrade_state_dict_named(self, state_dict, name):
        """Upgrade a (possibly old) state dict for new versions of fairseq."""
        if isinstance(self.embed_positions, SinusoidalPositionalEmbedding):
            weights_key = "{}.embed_positions.weights".format(name)
            if weights_key in state_dict:
                print("deleting {0}".format(weights_key))
                del state_dict[weights_key]
            state_dict[
                "{}.embed_positions._float_tensor".format(name)
            ] = torch.FloatTensor(1)
        for i in range(self.num_layers):
            # update layer norms
            self.layers[i].upgrade_state_dict_named(
                state_dict, "{}.layers.{}".format(name, i)
            )

        version_key = "{}.version".format(name)
        if utils.item(state_dict.get(version_key, torch.Tensor([1]))[0]) < 2:
            # earlier checkpoints did not normalize after the stack of layers
            self.layer_norm = None
            self.normalize = False
            state_dict[version_key] = torch.Tensor([1])
        return state_dict


class TransformerDecoder(FairseqIncrementalDecoder):
    """
    Transformer decoder consisting of *args.decoder_layers* layers. Each layer
    is a :class:`TransformerDecoderLayer`.

    Args:
        args (configargparse.Namespace): parsed command-line arguments
        dictionary (~fairseq.data.Dictionary): decoding dictionary
        embed_tokens (torch.nn.Embedding): output embedding
        no_encoder_attn (bool, optional): whether to attend to encoder outputs
            (default: False).
    """

    def __init__(self, args, dictionary, embed_tokens, no_encoder_attn=False):
        super().__init__(dictionary)
        self.register_buffer("version", torch.Tensor([3]))
        self._future_mask = torch.empty(0)

        self.dropout = args.dropout
        self.decoder_layerdrop = args.decoder_layerdrop
        self.share_input_output_embed = args.share_decoder_input_output_embed

        input_embed_dim = embed_tokens.embedding_dim
        embed_dim = args.decoder_embed_dim
        self.embed_dim = embed_dim
        self.output_embed_dim = args.decoder_output_dim

        self.padding_idx = embed_tokens.padding_idx
        self.max_target_positions = args.max_target_positions

        self.embed_tokens = embed_tokens

        self.embed_scale = 1.0 if args.no_scale_embedding else math.sqrt(embed_dim)

        self.project_in_dim = (
            Linear(input_embed_dim, embed_dim, bias=False)
            if embed_dim != input_embed_dim
            else None
        )

        self.embed_positions = (
            PositionalEmbedding(
                args.max_target_positions,
                embed_dim,
                self.padding_idx,
                learned=args.decoder_learned_pos,
            )
            if (not args.no_token_positional_embeddings and not args.no_decoder_positional_embeddings)
            else None
        )

        self.cross_self_attention = getattr(args, "cross_self_attention", False)
        self.layer_wise_attention = getattr(args, "layer_wise_attention", False)

        self.layers = nn.ModuleList([])
        self.layers.extend(
            [
                TransformerDecoderLayer(args, no_encoder_attn)
                for _ in range(args.decoder_layers)
            ]
        )
        self.num_layers = len(self.layers)

        self.adaptive_softmax = None

        self.project_out_dim = (
            Linear(embed_dim, self.output_embed_dim, bias=False)
            if embed_dim != self.output_embed_dim and not args.tie_adaptive_weights
            else None
        )

        if args.adaptive_softmax_cutoff is not None:
            self.adaptive_softmax = AdaptiveSoftmax(
                len(dictionary),
                self.output_embed_dim,
                options.eval_str_list(args.adaptive_softmax_cutoff, type=int),
                dropout=args.adaptive_softmax_dropout,
                adaptive_inputs=embed_tokens if args.tie_adaptive_weights else None,
                factor=args.adaptive_softmax_factor,
                tie_proj=args.tie_adaptive_proj,
            )
        elif not self.share_input_output_embed:
            self.embed_out = nn.Parameter(
                torch.Tensor(len(dictionary), self.output_embed_dim)
            )
            nn.init.normal_(self.embed_out, mean=0, std=self.output_embed_dim ** -0.5)

        if args.decoder_normalize_before and not getattr(
            args, "no_decoder_final_norm", False
        ):
            self.layer_norm = LayerNorm(embed_dim)
        else:
            self.layer_norm = None
        if getattr(args, "layernorm_embedding", False):
            self.layernorm_embedding = LayerNorm(embed_dim)
        else:
            self.layernorm_embedding = None

    def forward(
        self,
        prev_output_tokens,
        encoder_out: Optional[EncoderOut] = None,
        incremental_state: Optional[Dict[str, Dict[str, Optional[Tensor]]]] = None,
        features_only: bool = False,
        alignment_layer: Optional[int] = None,
        alignment_heads: Optional[int] = None,
        src_lengths: Optional[Any] = None,
        return_all_hiddens: bool = False,
    ):
        """
        Args:
            prev_output_tokens (LongTensor): previous decoder outputs of shape
                `(batch, tgt_len)`, for teacher forcing
            encoder_out (optional): output from the encoder, used for
                encoder-side attention
            incremental_state (dict): dictionary used for storing state during
                :ref:`Incremental decoding`
            features_only (bool, optional): only return features without
                applying output layer (default: False).

        Returns:
            tuple:
                - the decoder's output of shape `(batch, tgt_len, vocab)`
                - a dictionary with any model-specific outputs
        """
        x, extra = self.extract_features(
            prev_output_tokens,
            encoder_out=encoder_out,
            incremental_state=incremental_state,
            alignment_layer=alignment_layer,
            alignment_heads=alignment_heads,
        )
        if not features_only:
            x = self.output_layer(x)
        return x, extra

    def extract_features(
        self,
        prev_output_tokens,
        encoder_out: Optional[EncoderOut] = None,
        incremental_state: Optional[Dict[str, Dict[str, Optional[Tensor]]]] = None,
        full_context_alignment: bool = False,
        alignment_layer: Optional[int] = None,
        alignment_heads: Optional[int] = None,
    ):
        """
        Similar to *forward* but only return features.

        Includes several features from "Jointly Learning to Align and
        Translate with Transformer Models" (Garg et al., EMNLP 2019).

        Args:
            full_context_alignment (bool, optional): don't apply
                auto-regressive mask to self-attention (default: False).
            alignment_layer (int, optional): return mean alignment over
                heads at this layer (default: last layer).
            alignment_heads (int, optional): only average alignment over
                this many heads (default: all heads).

        Returns:
            tuple:
                - the decoder's features of shape `(batch, tgt_len, embed_dim)`
                - a dictionary with any model-specific outputs
        """
        if alignment_layer is None:
            alignment_layer = self.num_layers - 1

        # embed positions
        positions = (
            self.embed_positions(
                prev_output_tokens, incremental_state=incremental_state
            )
            if self.embed_positions is not None
            else None
        )

        if incremental_state is not None:
            prev_output_tokens = prev_output_tokens[:, -1:]
            if positions is not None:
                positions = positions[:, -1:]

        # embed tokens and positions
        x = self.embed_scale * self.embed_tokens(prev_output_tokens)

        if self.project_in_dim is not None:
            x = self.project_in_dim(x)

        if positions is not None:
            x += positions

        if self.layernorm_embedding is not None:
            x = self.layernorm_embedding(x)

        x = F.dropout(x, p=self.dropout, training=self.training)

        # B x T x C -> T x B x C
        x = x.transpose(0, 1)

        self_attn_padding_mask: Optional[Tensor] = None
        if self.cross_self_attention or prev_output_tokens.eq(self.padding_idx).any():
            self_attn_padding_mask = prev_output_tokens.eq(self.padding_idx)

        # decoder layers
        attn: Optional[Tensor] = None
        inner_states: List[Optional[Tensor]] = [x]
        for idx, layer in enumerate(self.layers):
            encoder_state: Optional[Tensor] = None
            if encoder_out is not None:
                if self.layer_wise_attention:
                    encoder_states = encoder_out.encoder_states
                    assert encoder_states is not None
                    encoder_state = encoder_states[idx]
                else:
                    encoder_state = encoder_out.encoder_out

            if incremental_state is None and not full_context_alignment:
                self_attn_mask = self.buffered_future_mask(x)
            else:
                self_attn_mask = None

            # add LayerDrop (see https://arxiv.org/abs/1909.11556 for description)
            dropout_probability = torch.empty(1).uniform_()
            if not self.training or (dropout_probability > self.decoder_layerdrop):
                x, layer_attn, _ = layer(
                    x,
                    encoder_state,
                    encoder_out.encoder_padding_mask
                    if encoder_out is not None
                    else None,
                    incremental_state,
                    self_attn_mask=self_attn_mask,
                    self_attn_padding_mask=self_attn_padding_mask,
                    need_attn=bool((idx == alignment_layer)),
                    need_head_weights=bool((idx == alignment_layer)),
                )
                inner_states.append(x)
                if layer_attn is not None and idx == alignment_layer:
                    attn = layer_attn.float().to(x)

        if attn is not None:
            if alignment_heads is not None:
                attn = attn[:alignment_heads]

            # average probabilities over heads
            attn = attn.mean(dim=0)

        if self.layer_norm is not None:
            x = self.layer_norm(x)

        # T x B x C -> B x T x C
        x = x.transpose(0, 1)

        if self.project_out_dim is not None:
            x = self.project_out_dim(x)

        return x, {"attn": [attn], "inner_states": inner_states}

    def output_layer(self, features):
        """Project features to the vocabulary size."""
        if self.adaptive_softmax is None:
            # project back to size of vocabulary
            if self.share_input_output_embed:
                return F.linear(features, self.embed_tokens.weight)
            else:
                return F.linear(features, self.embed_out)
        else:
            return features

    def max_positions(self):
        """Maximum output length supported by the decoder."""
        if self.embed_positions is None:
            return self.max_target_positions
        return min(self.max_target_positions, self.embed_positions.max_positions)

    def buffered_future_mask(self, tensor):
        dim = tensor.size(0)
        # self._future_mask.device != tensor.device is not working in TorchScript. This is a workaround.
        if (
            self._future_mask.size(0) == 0
            or (not self._future_mask.device == tensor.device)
            or self._future_mask.size(0) < dim
        ):
            self._future_mask = torch.triu(
                utils.fill_with_neg_inf(torch.zeros([dim, dim])), 1
            )
        self._future_mask = self._future_mask.to(tensor)
        return self._future_mask[:dim, :dim]

    def upgrade_state_dict_named(self, state_dict, name):
        """Upgrade a (possibly old) state dict for new versions of fairseq."""
        if isinstance(self.embed_positions, SinusoidalPositionalEmbedding):
            weights_key = "{}.embed_positions.weights".format(name)
            if weights_key in state_dict:
                del state_dict[weights_key]
            state_dict[
                "{}.embed_positions._float_tensor".format(name)
            ] = torch.FloatTensor(1)

        for i in range(self.num_layers):
            # update layer norms
            layer_norm_map = {
                "0": "self_attn_layer_norm",
                "1": "encoder_attn_layer_norm",
                "2": "final_layer_norm",
            }
            for old, new in layer_norm_map.items():
                for m in ("weight", "bias"):
                    k = "{}.layers.{}.layer_norms.{}.{}".format(name, i, old, m)
                    if k in state_dict:
                        state_dict[
                            "{}.layers.{}.{}.{}".format(name, i, new, m)
                        ] = state_dict[k]
                        del state_dict[k]

        version_key = "{}.version".format(name)
        if utils.item(state_dict.get(version_key, torch.Tensor([1]))[0]) <= 2:
            # earlier checkpoints did not normalize after the stack of layers
            self.layer_norm = None
            self.normalize = False
            state_dict[version_key] = torch.Tensor([1])

        return state_dict


def Embedding(num_embeddings, embedding_dim, padding_idx):
    m = nn.Embedding(num_embeddings, embedding_dim, padding_idx=padding_idx)
    nn.init.normal_(m.weight, mean=0, std=embedding_dim ** -0.5)
    nn.init.constant_(m.weight[padding_idx], 0)
    return m


def Linear(in_features, out_features, bias=True):
    m = nn.Linear(in_features, out_features, bias)
    nn.init.xavier_uniform_(m.weight)
    if bias:
        nn.init.constant_(m.bias, 0.0)
    return m


@register_model_architecture("transformer", "transformer")
def base_architecture(args):
    args.encoder_embed_path = getattr(args, "encoder_embed_path", None)
    args.encoder_embed_dim = getattr(args, "encoder_embed_dim", 512)
    args.encoder_ffn_embed_dim = getattr(args, "encoder_ffn_embed_dim", 2048)
    args.encoder_layers = getattr(args, "encoder_layers", 6)
    args.encoder_attention_heads = getattr(args, "encoder_attention_heads", 8)
    args.encoder_normalize_before = getattr(args, "encoder_normalize_before", False)
    args.encoder_learned_pos = getattr(args, "encoder_learned_pos", False)
    args.decoder_embed_path = getattr(args, "decoder_embed_path", None)
    args.decoder_embed_dim = getattr(args, "decoder_embed_dim", args.encoder_embed_dim)
    args.decoder_ffn_embed_dim = getattr(
        args, "decoder_ffn_embed_dim", args.encoder_ffn_embed_dim
    )
    args.decoder_layers = getattr(args, "decoder_layers", 6)
    args.decoder_attention_heads = getattr(args, "decoder_attention_heads", 8)
    args.decoder_normalize_before = getattr(args, "decoder_normalize_before", False)
    args.decoder_learned_pos = getattr(args, "decoder_learned_pos", False)
    args.attention_dropout = getattr(args, "attention_dropout", 0.0)
    args.activation_dropout = getattr(args, "activation_dropout", 0.0)
    args.activation_fn = getattr(args, "activation_fn", "relu")
    args.dropout = getattr(args, "dropout", 0.1)
    args.adaptive_softmax_cutoff = getattr(args, "adaptive_softmax_cutoff", None)
    args.adaptive_softmax_dropout = getattr(args, "adaptive_softmax_dropout", 0)
    args.share_decoder_input_output_embed = getattr(
        args, "share_decoder_input_output_embed", False
    )
    args.share_all_embeddings = getattr(args, "share_all_embeddings", False)
    args.no_token_positional_embeddings = getattr(
        args, "no_token_positional_embeddings", False
    )
    args.no_decoder_positional_embeddings = getattr(
        args, "no_decoder_positional_embeddings", False
    )
    args.adaptive_input = getattr(args, "adaptive_input", False)
    args.no_cross_attention = getattr(args, "no_cross_attention", False)
    args.cross_self_attention = getattr(args, "cross_self_attention", False)
    args.layer_wise_attention = getattr(args, "layer_wise_attention", False)

    args.decoder_output_dim = getattr(
        args, "decoder_output_dim", args.decoder_embed_dim
    )
    args.decoder_input_dim = getattr(args, "decoder_input_dim", args.decoder_embed_dim)

    args.no_scale_embedding = getattr(args, "no_scale_embedding", False)
    args.layernorm_embedding = getattr(args, "layernorm_embedding", False)


@register_model_architecture("transformer", "transformer_iwslt_de_en")
def transformer_iwslt_de_en(args):
    args.encoder_embed_dim = getattr(args, "encoder_embed_dim", 512)
    args.encoder_ffn_embed_dim = getattr(args, "encoder_ffn_embed_dim", 1024)
    args.encoder_attention_heads = getattr(args, "encoder_attention_heads", 4)
    args.encoder_layers = getattr(args, "encoder_layers", 6)
    args.decoder_embed_dim = getattr(args, "decoder_embed_dim", 512)
    args.decoder_ffn_embed_dim = getattr(args, "decoder_ffn_embed_dim", 1024)
    args.decoder_attention_heads = getattr(args, "decoder_attention_heads", 4)
    args.decoder_layers = getattr(args, "decoder_layers", 6)
    base_architecture(args)


@register_model_architecture("transformer", "transformer_wmt_en_de")
def transformer_wmt_en_de(args):
    base_architecture(args)


# parameters used in the "Attention Is All You Need" paper (Vaswani et al., 2017)
@register_model_architecture("transformer", "transformer_vaswani_wmt_en_de_big")
def transformer_vaswani_wmt_en_de_big(args):
    args.encoder_embed_dim = getattr(args, "encoder_embed_dim", 1024)
    args.encoder_ffn_embed_dim = getattr(args, "encoder_ffn_embed_dim", 4096)
    args.encoder_attention_heads = getattr(args, "encoder_attention_heads", 16)
    args.encoder_normalize_before = getattr(args, "encoder_normalize_before", False)
    args.decoder_embed_dim = getattr(args, "decoder_embed_dim", 1024)
    args.decoder_ffn_embed_dim = getattr(args, "decoder_ffn_embed_dim", 4096)
    args.decoder_attention_heads = getattr(args, "decoder_attention_heads", 16)
    args.dropout = getattr(args, "dropout", 0.3)
    base_architecture(args)


@register_model_architecture("transformer", "transformer_vaswani_wmt_en_fr_big")
def transformer_vaswani_wmt_en_fr_big(args):
    args.dropout = getattr(args, "dropout", 0.1)
    transformer_vaswani_wmt_en_de_big(args)


@register_model_architecture("transformer", "transformer_wmt_en_de_big")
def transformer_wmt_en_de_big(args):
    args.attention_dropout = getattr(args, "attention_dropout", 0.1)
    transformer_vaswani_wmt_en_de_big(args)


# default parameters used in tensor2tensor implementation
@register_model_architecture("transformer", "transformer_wmt_en_de_big_t2t")
def transformer_wmt_en_de_big_t2t(args):
    args.encoder_normalize_before = getattr(args, "encoder_normalize_before", True)
    args.decoder_normalize_before = getattr(args, "decoder_normalize_before", True)
    args.attention_dropout = getattr(args, "attention_dropout", 0.1)
    args.activation_dropout = getattr(args, "activation_dropout", 0.1)
    transformer_vaswani_wmt_en_de_big(args)


@register_model_architecture("transformer_align", "transformer_align")
def transformer_align(args):
    args.alignment_heads = getattr(args, "alignment_heads", 1)
    args.alignment_layer = getattr(args, "alignment_layer", 4)
    args.full_context_alignment = getattr(args, "full_context_alignment", False)
    base_architecture(args)


@register_model_architecture("transformer_align", "transformer_wmt_en_de_big_align")
def transformer_wmt_en_de_big_align(args):
    args.alignment_heads = getattr(args, "alignment_heads", 1)
    args.alignment_layer = getattr(args, "alignment_layer", 4)
    transformer_wmt_en_de_big(args)
